from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class Colour_SDS(_message.Message):
    __slots__ = ["name", "shape", "values", "values_double"]
    NAME_FIELD_NUMBER: _ClassVar[int]
    SHAPE_FIELD_NUMBER: _ClassVar[int]
    VALUES_DOUBLE_FIELD_NUMBER: _ClassVar[int]
    VALUES_FIELD_NUMBER: _ClassVar[int]
    name: str
    shape: Spectral_Shape
    values: _containers.RepeatedScalarFieldContainer[float]
    values_double: _containers.RepeatedScalarFieldContainer[float]
    def __init__(self, shape: _Optional[_Union[Spectral_Shape, _Mapping]] = ..., values_double: _Optional[_Iterable[float]] = ..., values: _Optional[_Iterable[float]] = ..., name: _Optional[str] = ...) -> None: ...

class Measurement(_message.Message):
    __slots__ = ["XYZ", "anc_data", "cct", "dominant_wl", "exposure", "power", "spd", "spectrometer_id", "time", "xy"]
    ANC_DATA_FIELD_NUMBER: _ClassVar[int]
    CCT_FIELD_NUMBER: _ClassVar[int]
    DOMINANT_WL_FIELD_NUMBER: _ClassVar[int]
    EXPOSURE_FIELD_NUMBER: _ClassVar[int]
    POWER_FIELD_NUMBER: _ClassVar[int]
    SPD_FIELD_NUMBER: _ClassVar[int]
    SPECTROMETER_ID_FIELD_NUMBER: _ClassVar[int]
    TIME_FIELD_NUMBER: _ClassVar[int]
    XYZ: XYZ_value
    XYZ_FIELD_NUMBER: _ClassVar[int]
    XY_FIELD_NUMBER: _ClassVar[int]
    anc_data: bytes
    cct: cct_value
    dominant_wl: float
    exposure: float
    power: float
    spd: Colour_SDS
    spectrometer_id: str
    time: Timestamp
    xy: xy_value
    def __init__(self, spd: _Optional[_Union[Colour_SDS, _Mapping]] = ..., exposure: _Optional[float] = ..., XYZ: _Optional[_Union[XYZ_value, _Mapping]] = ..., xy: _Optional[_Union[xy_value, _Mapping]] = ..., cct: _Optional[_Union[cct_value, _Mapping]] = ..., dominant_wl: _Optional[float] = ..., power: _Optional[float] = ..., spectrometer_id: _Optional[str] = ..., time: _Optional[_Union[Timestamp, _Mapping]] = ..., anc_data: _Optional[bytes] = ...) -> None: ...

class Measurement_List(_message.Message):
    __slots__ = ["ancillary", "author", "location", "measurements", "notes", "order", "software", "test_colors"]
    class TestColor(_message.Message):
        __slots__ = ["c", "f"]
        C_FIELD_NUMBER: _ClassVar[int]
        F_FIELD_NUMBER: _ClassVar[int]
        c: _containers.RepeatedScalarFieldContainer[int]
        f: _containers.RepeatedScalarFieldContainer[float]
        def __init__(self, c: _Optional[_Iterable[int]] = ..., f: _Optional[_Iterable[float]] = ...) -> None: ...
    ANCILLARY_FIELD_NUMBER: _ClassVar[int]
    AUTHOR_FIELD_NUMBER: _ClassVar[int]
    LOCATION_FIELD_NUMBER: _ClassVar[int]
    MEASUREMENTS_FIELD_NUMBER: _ClassVar[int]
    NOTES_FIELD_NUMBER: _ClassVar[int]
    ORDER_FIELD_NUMBER: _ClassVar[int]
    SOFTWARE_FIELD_NUMBER: _ClassVar[int]
    TEST_COLORS_FIELD_NUMBER: _ClassVar[int]
    ancillary: bytes
    author: str
    location: str
    measurements: _containers.RepeatedCompositeFieldContainer[Measurement]
    notes: str
    order: _containers.RepeatedScalarFieldContainer[int]
    software: str
    test_colors: _containers.RepeatedCompositeFieldContainer[Measurement_List.TestColor]
    def __init__(self, measurements: _Optional[_Iterable[_Union[Measurement, _Mapping]]] = ..., test_colors: _Optional[_Iterable[_Union[Measurement_List.TestColor, _Mapping]]] = ..., order: _Optional[_Iterable[int]] = ..., notes: _Optional[str] = ..., author: _Optional[str] = ..., location: _Optional[str] = ..., software: _Optional[str] = ..., ancillary: _Optional[bytes] = ...) -> None: ...

class Spectral_Shape(_message.Message):
    __slots__ = ["end", "start", "step"]
    END_FIELD_NUMBER: _ClassVar[int]
    START_FIELD_NUMBER: _ClassVar[int]
    STEP_FIELD_NUMBER: _ClassVar[int]
    end: float
    start: float
    step: float
    def __init__(self, start: _Optional[float] = ..., end: _Optional[float] = ..., step: _Optional[float] = ...) -> None: ...

class Timestamp(_message.Message):
    __slots__ = ["timestr"]
    TIMESTR_FIELD_NUMBER: _ClassVar[int]
    timestr: str
    def __init__(self, timestr: _Optional[str] = ...) -> None: ...

class XYZ_value(_message.Message):
    __slots__ = ["X", "Y", "Z"]
    X: float
    X_FIELD_NUMBER: _ClassVar[int]
    Y: float
    Y_FIELD_NUMBER: _ClassVar[int]
    Z: float
    Z_FIELD_NUMBER: _ClassVar[int]
    def __init__(self, X: _Optional[float] = ..., Y: _Optional[float] = ..., Z: _Optional[float] = ...) -> None: ...

class cct_value(_message.Message):
    __slots__ = ["cct", "duv"]
    CCT_FIELD_NUMBER: _ClassVar[int]
    DUV_FIELD_NUMBER: _ClassVar[int]
    cct: float
    duv: float
    def __init__(self, cct: _Optional[float] = ..., duv: _Optional[float] = ...) -> None: ...

class xy_value(_message.Message):
    __slots__ = ["x", "y"]
    X_FIELD_NUMBER: _ClassVar[int]
    Y_FIELD_NUMBER: _ClassVar[int]
    x: float
    y: float
    def __init__(self, x: _Optional[float] = ..., y: _Optional[float] = ...) -> None: ...



from .CRSpectrometer import CRSpectrometer
from .CR_Definitions import (
    CommandError,
    CommandResponse,
    InstrumentType,
    MeasurementSpeed,
    Model,
    ResponseCode,
)

__all__ = [
    "CRSpectrometer",
    "CommandError",
    "CommandResponse",
    "InstrumentType",
    "MeasurementSpeed",
    "Model",
    "ResponseCode",
]

__author__ = "Tucker Downs"
__copyright__ = "Copyright 2022 Specio Developers"
__license__ = "MIT License - https://github.com/tjdcs/specio/blob/main/LICENSE.md"
__maintainer__ = "Tucker Downs"
__email__ = "tucker@tuckerd.info"
__status__ = "Development"



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: measurements.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x12measurements.proto\x12\x0especio.protoio\":\n\x0eSpectral_Shape\x12\r\n\x05start\x18\x01 \x01(\x01\x12\x0b\n\x03\x65nd\x18\x02 \x01(\x01\x12\x0c\n\x04step\x18\x03 \x01(\x01\"~\n\nColour_SDS\x12-\n\x05shape\x18\x01 \x01(\x0b\x32\x1e.specio.protoio.Spectral_Shape\x12\x15\n\rvalues_double\x18\x02 \x03(\x01\x12\x0e\n\x06values\x18\x04 \x03(\x02\x12\x11\n\x04name\x18\x03 \x01(\tH\x00\x88\x01\x01\x42\x07\n\x05_name\",\n\tXYZ_value\x12\t\n\x01X\x18\x01 \x01(\x01\x12\t\n\x01Y\x18\x02 \x01(\x01\x12\t\n\x01Z\x18\x03 \x01(\x01\" \n\x08xy_value\x12\t\n\x01x\x18\x01 \x01(\x01\x12\t\n\x01y\x18\x02 \x01(\x01\"%\n\tcct_value\x12\x0b\n\x03\x63\x63t\x18\x01 \x01(\x01\x12\x0b\n\x03\x64uv\x18\x02 \x01(\x01\"\x1c\n\tTimestamp\x12\x0f\n\x07timestr\x18\x01 \x01(\t\"\xc8\x02\n\x0bMeasurement\x12\'\n\x03spd\x18\x01 \x01(\x0b\x32\x1a.specio.protoio.Colour_SDS\x12\x10\n\x08\x65xposure\x18\x02 \x01(\x01\x12&\n\x03XYZ\x18\x03 \x01(\x0b\x32\x19.specio.protoio.XYZ_value\x12$\n\x02xy\x18\x04 \x01(\x0b\x32\x18.specio.protoio.xy_value\x12&\n\x03\x63\x63t\x18\x05 \x01(\x0b\x32\x19.specio.protoio.cct_value\x12\x13\n\x0b\x64ominant_wl\x18\x06 \x01(\x01\x12\r\n\x05power\x18\x07 \x01(\x01\x12\x17\n\x0fspectrometer_id\x18\x08 \x01(\t\x12\'\n\x04time\x18\t \x01(\x0b\x32\x19.specio.protoio.Timestamp\x12\x15\n\x08\x61nc_data\x18\x10 \x01(\x0cH\x00\x88\x01\x01\x42\x0b\n\t_anc_data\"\x8e\x02\n\x10Measurement_List\x12\x31\n\x0cmeasurements\x18\x01 \x03(\x0b\x32\x1b.specio.protoio.Measurement\x12?\n\x0btest_colors\x18\x03 \x03(\x0b\x32*.specio.protoio.Measurement_List.TestColor\x12\r\n\x05order\x18\x02 \x03(\x05\x12\r\n\x05notes\x18\x10 \x01(\t\x12\x0e\n\x06\x61uthor\x18\x11 \x01(\t\x12\x10\n\x08location\x18\x12 \x01(\t\x12\x10\n\x08software\x18\x13 \x01(\t\x12\x11\n\tancillary\x18\x14 \x01(\x0c\x1a!\n\tTestColor\x12\t\n\x01\x63\x18\x01 \x03(\r\x12\t\n\x01\x66\x18\x02 \x03(\x02\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'measurements_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _SPECTRAL_SHAPE._serialized_start=38
  _SPECTRAL_SHAPE._serialized_end=96
  _COLOUR_SDS._serialized_start=98
  _COLOUR_SDS._serialized_end=224
  _XYZ_VALUE._serialized_start=226
  _XYZ_VALUE._serialized_end=270
  _XY_VALUE._serialized_start=272
  _XY_VALUE._serialized_end=304
  _CCT_VALUE._serialized_start=306
  _CCT_VALUE._serialized_end=343
  _TIMESTAMP._serialized_start=345
  _TIMESTAMP._serialized_end=373
  _MEASUREMENT._serialized_start=376
  _MEASUREMENT._serialized_end=704
  _MEASUREMENT_LIST._serialized_start=707
  _MEASUREMENT_LIST._serialized_end=977
  _MEASUREMENT_LIST_TESTCOLOR._serialized_start=944
  _MEASUREMENT_LIST_TESTCOLOR._serialized_end=977
# @@protoc_insertion_point(module_scope)



"""
Implement support for operations with CR300
"""
import logging
import platform
import re
import time

import serial.tools.list_ports
from colour import SpectralDistribution, SpectralShape

from specio.common import RawMeasurement
from specio.spectrometer import SpecRadiometer

from specio.ColorimetryResearch.CR_Definitions import (
    CommandError,
    CommandResponse,
    InstrumentType,
    MeasurementSpeed,
    Model,
    ResponseCode,
    ResponseType,
)

__author__ = "Tucker Downs"
__copyright__ = "Copyright 2022 Specio Developers"
__license__ = "MIT License - https://github.com/tjdcs/specio/blob/main/LICENSE.md"
__maintainer__ = "Tucker Downs"
__email__ = "tucker@tuckerd.info"
__status__ = "Development"

DEFAULT_TIMEOUT = 0.01


class CRSpectrometer(SpecRadiometer):
    """
    RAII Colorimetry Research Spectroradiometers.

    In particular this implementation was developed and tested with the CR300,
    but it should work for CR250 or other spectrometers from CR.
    """

    def __initialize_connection__(self, device: str | None = None) -> str:
        if device is None:
            if platform.system() == "Darwin":
                portlist = list(serial.tools.list_ports.grep("usbmodem"))
            elif platform.system() == "Windows":
                portlist = list(serial.tools.list_ports.grep("Colorimetry"))
            elif platform.system() == "Unix":
                raise NotImplementedError("CR discovery is not implemented for Unix")
            elif platform.system() == "Linux":
                portlist = list(serial.tools.list_ports.grep("ACM"))

            if len(portlist) == 0:
                raise OSError("Could not find any candidates for ColorimetryResearch")

            if len(portlist) > 1:
                raise NotImplementedError(
                    "Too many port candidates found, only a single port may be auto discovered."
                )

            device = portlist[0].device

        self.__port = serial.Serial(device, timeout=DEFAULT_TIMEOUT, baudrate=115200)

    def __init__(
        self,
        device: str | None = None,
        speed: MeasurementSpeed = MeasurementSpeed.NORMAL,
    ):
        """
        Construct CR Controller Obj
        """
        self.__last_cmd_time: float = 0
        self.__initialize_connection__(device=device)

        if self.instrument_type is not InstrumentType.SPECTRORADIOMETER:
            raise RuntimeError("The connected-to device was not a spectroradiometer.")

        self.measurement_speed = speed

        super().__init__(
            manufacturer="Colorimetry Research",
            model=self.model,
            serial_num=self.serial_number,
        )

    @property
    def firmware(self):
        if not hasattr(self, "_firmware") or self._firmware is None:
            response = self.__write_cmd("RC Firmware")
            self._firmware = response.arguments[0]
        return self._firmware

    @property
    def measurement_speed(self) -> MeasurementSpeed:
        response = self.__write_cmd("SM ExposureMode 0")
        response = self.__write_cmd("RS Speed")
        self._measurement_speed = MeasurementSpeed(response.arguments[0].lower())
        return self._measurement_speed

    @measurement_speed.setter
    def measurement_speed(self, speed: MeasurementSpeed):
        t = self.__write_cmd(f"SM Speed {speed.values[0]}")
        self._measurement_speed = speed

    @property
    def aperture(self):
        """
        Get spectrometer aperture value
        """
        if not hasattr(self, "_aperture") or self._aperture is None:
            response = self.__write_cmd("RS Aperture")
            self._aperture = response.arguments[0]
        return self._aperture

    @property
    def serial_number(self):
        if not hasattr(self, "_sn") or self._sn is None:
            response = self.__write_cmd("RC ID")
            self._sn = response.arguments[0]
        return self._sn

    @property
    def model(self):
        if not hasattr(self, "_model") or self._model is None:
            response = self.__write_cmd("RC Model")
            self._model = response.arguments[0]
        return self._model

    @property
    def instrument_type(self):
        """
        Check that the connected device is a spectrometer
        """
        if not hasattr(self, "_instrument_type") or self._instrument_type is None:
            response = self.__write_cmd("RC InstrumentType")
            i_type = InstrumentType(response.arguments[0])
            self._instrument_type = i_type

        return self._instrument_type

    def __clear_buffer(self):
        """
        Clear input buffer
        """
        self.__port.readall()

    def __write_cmd(self, command: str) -> CommandResponse:
        """
        Write cmd to serial port
        """
        log = logging.getLogger("specio.CR")
        log.debug(f"Sending CMD: {command}")

        command = (command + "\n").encode()
        self.__clear_buffer()
        if self.__last_cmd_time + DEFAULT_TIMEOUT > time.time():
            time.sleep(
                max(self.__last_cmd_time + DEFAULT_TIMEOUT + 0.001 - time.time(), 0)
            )

        self.__port.write(command)
        self.__last_cmd_time = time.time()

        while self.__port.in_waiting < 1:
            time.sleep(0.01)

        response = self.__port.readline()

        response = self.__parse_response(response)

        if response.type == ResponseType.error:
            raise CommandError(response, response.arguments[0])
        else:
            return response

    def __parse_response(self, response: str | bytes) -> CommandResponse:
        """
        Parse CR response string
        """
        if type(response) is bytes:
            response = response.decode()
        response = response.strip().split(":")

        args = []
        if response[3].isnumeric() and int(response[3]) > 0 and self.__port.inWaiting():
            for line_num in range(int(response[3])):
                response = self.__port.readline()
                args.append(response)
        else:
            args = response[3:]

        return CommandResponse(
            type=ResponseType(response[0]),
            code=ResponseCode(int(response[1])),
            description=response[2],
            arguments=args,
        )

    def _raw_measure(self) -> SpectralDistribution:
        """
        Make spectral measurement with CR
        """
        response = self.__write_cmd("M")
        response = self.__write_cmd("RM Spectrum")

        args = response.arguments[0].split(",")
        shape = SpectralShape(
            start=float(args[0]), end=float(args[1]), interval=float(args[2])
        )

        data = self.__port.readall()
        data = [float(d) for d in data.decode().splitlines()]

        exposure = self.__write_cmd("RM Exposure").arguments[0]
        exMatch = re.match("\d*\.?\d*", exposure)
        if exMatch:
            exposure = float(exMatch.group()) / 1000
        else:
            exposure = -1

        return RawMeasurement(
            spd=SpectralDistribution(data=data, domain=shape),
            spectrometer_id=self.readable_id,
            exposure=exposure,
        )


if __name__ == "__main__":
    cr = CRSpectrometer()

    print(f"Aperture: {cr.aperture}")
    print(f"Firmware: {cr.firmware}")

    print(f"Measurement Speed: {cr.measurement_speed}")
    cr.measurement_speed = MeasurementSpeed.FAST_2X
    print(f"Measurement Speed: {cr.measurement_speed}")

    print(f"Instrument Type: {cr.instrument_type}")

    pass



from datetime import datetime

import numpy as np
from colour import sd_multi_leds

from specio.common import Measurement, RawMeasurement

__author__ = "Tucker Downs"
__copyright__ = "Copyright 2022 Specio Developers"
__license__ = "MIT License - https://github.com/tjdcs/specio/blob/main/LICENSE.md"
__maintainer__ = "Tucker Downs"
__email__ = "tucker@tuckerd.info"
__status__ = "Development"


class SpecRadiometer:
    def __init__(
        self,
        model: str = "Virtual Spectrometer",
        manufacturer: str = "Specio",
        serial_num: str | None = None,
    ):
        self.manufacturer: str = manufacturer
        self._model: str = model
        self._serial_number: str | None = serial_num

    @property
    def model(self):
        return self._model

    @property
    def serial_number(self):
        return self._serial_number

    @property
    def readable_id(self) -> str:
        return f"{self.model} - {self.serial_number}"

    def measure(self) -> Measurement:
        _rm = self._raw_measure()
        _measurement = Measurement(_rm)
        return _measurement

    def _raw_measure(self) -> RawMeasurement:
        peaks = np.random.randint([460, 510, 600], [480, 570, 690], 3)
        widths = np.random.randint(40, 80, 3)
        powers = np.random.randint(10, 40, 3) / 1000
        spd = sd_multi_leds(
            peak_wavelengths=peaks, fwhm=widths, peak_power_ratios=powers
        )

        _measurement = RawMeasurement(
            spd=spd,
            exposure=0.1,
            spectrometer_id="Virtual Spectrometer",
        )
        return _measurement


if __name__ == "__main__":
    t = SpecRadiometer()
    m = t.measure()
    print(m)



import textwrap
from dataclasses import dataclass
from datetime import datetime

from typing import Any, Tuple
import colour

import numpy as np
from colour import (
    SpectralDistribution,
    SpectralShape,
    XYZ_to_xy,
    dominant_wavelength,
    sd_to_XYZ,
    uv_to_CCT,
    xy_to_UCS_uv,
)
from numpy import byte, ndarray, power
from specio.protobuf import measurements_pb2

__author__ = "Tucker Downs"
__copyright__ = "Copyright 2022 Specio Developers"
__license__ = "MIT License - https://github.com/tjdcs/specio/blob/main/LICENSE.md"
__maintainer__ = "Tucker Downs"
__email__ = "tucker@tuckerd.info"
__status__ = "Development"


@dataclass
class RawMeasurement:
    spd: SpectralDistribution
    exposure: float
    spectrometer_id: str
    anc_data: None | Any = None


@dataclass
class Measurement:
    def _from_protobuf(self, data: bytes | measurements_pb2.Measurement):
        m: measurements_pb2.Measurement
        if type(data) == bytes:
            m = measurements_pb2.Measurement()
            m.ParseFromString(data)
        else:
            m = data

        self.spd = SpectralDistribution(
            data=m.spd.values,
            domain=SpectralShape(
                start=m.spd.shape.start, end=m.spd.shape.end, interval=m.spd.shape.step
            ),
            name=m.spd.name,
        )
        self.exposure = m.exposure
        self.spectrometer_id = m.spectrometer_id
        self.XYZ = np.array([m.XYZ.X, m.XYZ.Y, m.XYZ.Z])
        self.xy = np.array([m.xy.x, m.xy.y])
        self.cct = m.cct.cct
        self.duv = m.cct.duv
        self.dominant_wl = m.dominant_wl
        self.power = m.power
        self.time = datetime.fromisoformat(m.time.timestr)
        self.anc_data = None

    def to_protobuf(self) -> Tuple[bytes, measurements_pb2.Measurement]:
        m = measurements_pb2.Measurement()
        m.exposure = self.exposure
        m.dominant_wl = self.dominant_wl
        m.power = self.power
        m.spectrometer_id = self.spectrometer_id

        m.time.timestr = self.time.isoformat()

        m.spd.values[:] = self.spd.values.tolist()
        m.spd.shape.start = self.spd.shape.start
        m.spd.shape.end = self.spd.shape.end
        m.spd.shape.step = self.spd.shape.interval
        m.spd.name = self.spd.name

        m.XYZ.X = self.XYZ[0]
        m.XYZ.Y = self.XYZ[1]
        m.XYZ.Z = self.XYZ[2]

        m.xy.x = self.xy[0]
        m.xy.y = self.xy[1]

        m.cct.cct = self.cct
        m.cct.duv = self.duv

        if self.anc_data is not None:
            pass

        return (m.SerializeToString(), m)

    def __init__(
        self,
        raw_meas: (None | RawMeasurement | bytes | measurements_pb2.Measurement) = None,
    ):
        if type(raw_meas) is bytes or type(raw_meas) == measurements_pb2.Measurement:
            self._from_protobuf(raw_meas)
            return

        if raw_meas is None:
            self.spd = colour.sd_multi_leds(
                np.random.randint(400, 700, size=3), np.random.randint(40, 150, size=3)
            )
            self.spd.values / 1000
            self.exposure = 1
            self.spectrometer_id = "Virtual Spectrometer"
            self.anc_data = None
        else:
            self.spd = raw_meas.spd
            self.exposure = raw_meas.exposure
            self.spectrometer_id = raw_meas.spectrometer_id
            self.anc_data = raw_meas.anc_data

        self.XYZ = sd_to_XYZ(self.spd, k=6.83)  # TODO colour-science > 0.4.2 k=683
        self.xy = XYZ_to_xy(self.XYZ)
        _cct = uv_to_CCT(xy_to_UCS_uv(self.xy))
        self.cct = _cct[0]
        self.duv = _cct[1]
        self.dominant_wl = dominant_wavelength(self.xy, [1 / 3, 1 / 3])[0]
        self.power = self.spd.values.sum() * 1000
        self.time = datetime.now()

    spd: SpectralDistribution
    exposure: float
    spectrometer_id: str
    XYZ: ndarray
    xy: ndarray
    cct: ndarray
    duv: ndarray
    dominant_wl: float
    power: float
    anc_data: None | Any
    time: datetime

    def __str__(self) -> str:
        return textwrap.dedent(
            f"""
            Spectral Measurement - {self.spectrometer_id}:
                time: {self.time}
                XYZ: {np.array2string(self.XYZ, formatter={'float_kind':lambda x: "%.2f" % x})}
                CCT: {self.cct:.0f} ± {self.duv:.5f}
                Dominant WL: {self.dominant_wl:.1f}
                Exposure: {self.exposure:.3f}
            """
        )


if __name__ == "__main__":
    m = Measurement()
    str = m.to_protobuf()[0]
    print(str)
    m2 = Measurement(str)
    pass



from typing import Any

from aenum import Enum, MultiValueEnum
from attr import dataclass

__author__ = "Tucker Downs"
__copyright__ = "Copyright 2022 Specio Developers"
__license__ = "MIT License - https://github.com/tjdcs/specio/blob/main/LICENSE.md"
__maintainer__ = "Tucker Downs"
__email__ = "tucker@tuckerd.info"
__status__ = "Development"


class InstrumentType(MultiValueEnum):
    """
    Identifies the type of instrument
    =====
    0: Photometer
    1: Colorimeter
    2: Spectroradiometer
    """

    PHOTOMETER = 0, "0"
    COLORIMETER = 1, "1"
    SPECTRORADIOMETER = 2, "2"


class MeasurementSpeed(MultiValueEnum):
    """
    Controls the measurement speed when the CR Exposure Mode is set to "auto"
    """

    SLOW = 0, "0", "slow"
    NORMAL = 1, "1", "normal"
    FAST = 2, "2", "fast"
    FAST_2X = 3, "3", "2x fast"


class Model(Enum):
    """
    Identifies the CR model
    =====
    Values are based on the response to "RC Model" command.
    """

    CR300 = "CR-300"
    CR250 = "CR-250"


class ResponseType(Enum):
    """
    Identifies the success / failure of any CR command.
    """

    error = "ER"
    ok = "OK"


class ResponseCode(MultiValueEnum):
    """
    Error codes for Colorimetry Research CR-300 and other model spectrometers
    =====
    """

    OK = 0
    INVALID_COMMAND = -500

    TOO_DARK = 100
    CANT_SYNC_CONST = 101
    CANT_SYNC_AUTO = 102
    SYNC_TOO_LOW = 103

    INVALID_SYNC_MODE = -300, -521
    INVALID_SYNC_PERIOD = -301
    CANT_SYNC_TO_LIGHT = -302

    LIGHT_INTENSITY_FLUCTUATION = -303
    LIGHT_INTENSITY_TOO_LOW = -304
    LIGHT_INTENSITY_UNMEASURABLE = -305
    LIGHT_INTENSITY_TOO_HIGH = -306

    HARDWARE_MALFUNCTION = -331
    MATRIX_VERSION_MISMATCH = -332
    INVALID_MATRIX_INDEX = -333

    NO_CIE_TABLES = -334
    NO_CMF_TABLES = -335

    NO_MATRIX_FOR_ID = -336

    DUPLICATE_FILTER_SELECTION = -505
    NO_ACCESSORY_FOR_INDEX = -506
    NO_FILTER_FOR_INDEX = -507
    INDEX_NOT_VALID_ACCESSORY = -508
    INDEX_NOT_VALID_FILTER = -509, -510, -511

    INVALID_RANGE_MODE = -512
    INVALID_RANGE_INDEX = -513
    INVALID_EXPOSURE_MULTIPLIER = -514

    INDEX_DOESNT_SELECT_APERTURE = -515

    INVALID_EXPOSURE_MODE = -518
    INVALID_EXPOSURE_VALUE = -519

    INVALID_SYNC_FREQUENCY = -522

    INVALID_MATRIX_MODE = -552
    INVALID_MATRIX_ID = -553
    INVALID_MATRIX_NAME = -555

    ERRROR_SAVING_MATRIX_FLASH = -559
    INVALID_USER_CALIBRATION_MODE = -560

    RESERVED = -999

    @classmethod
    def _missing_(cls, _: object) -> Any:
        """
        Set the default error code to RESERVED
        =======================================
        """
        return cls.RESERVED


@dataclass
class CommandResponse:
    """
    Response data from Colorimetry Research
    """

    type: ResponseCode
    code: ResponseCode
    description: str
    arguments: list[str]


class CommandError(Exception):
    """
    Describes an issue with sending a command to CR
    """

    def __init__(self, response: CommandResponse, *args: object) -> None:
        self.response = response
        super().__init__(*args)



"""
Specio
======
Provides support for interacting with various hardware spectrometers.
"""
__version__ = "0.2.0"

from .common import Measurement, RawMeasurement
from .spectrometer import SpecRadiometer

__all__ = [
    "Measurement",
    "RawMeasurement",
    "SpecRadiometer",
]


__author__ = "Tucker Downs"
__copyright__ = "Copyright 2022 Specio Developers"
__license__ = "MIT License - https://github.com/tjdcs/specio/blob/main/LICENSE.md"
__maintainer__ = "Tucker Downs"
__email__ = "tucker@tuckerd.info"
__status__ = "Development"


def config__specio_logger() -> None:
    """
    Configure the default specio logger "specio"
    """
    import logging

    log = logging.getLogger("specio")

    stream_handler = logging.StreamHandler()
    stream_handler.formatter = logging.Formatter(
        fmt="[%(asctime)s.%(msecs)d] %(levelname)s - %(name)s %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )

    log.addHandler(stream_handler)
    log.setLevel("DEBUG")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['specio', 'specio.ColorimetryResearch', 'specio.protobuf']

package_data = \
{'': ['*']}

install_requires = \
['aenum>=3.1.11,<4.0.0',
 'colour-science>=0.4.1,<0.5.0',
 'numpy>=1.23.1,<2.0.0',
 'protobuf>=4.21.8,<5.0.0',
 'pyserial>=3.5,<4.0']

setup_kwargs = {
    'name': 'colour-specio',
    'version': '0.2.7',
    'description': 'Instrument control for spectrometers',
    'long_description': '# Specio\n\nSpecio is a python library for interacting with spectrometers. Currently only\nthe Colorimetry Research family is supported, and particularly this library is\ntested and maintained with a CR300.\n\nThis library also provides a virtual spectrometer which provides semi-random\nSPDs as measurements.\n\n## Usage\n\nSee Examples Folder\n',
    'author': 'Tucker',
    'author_email': 'tucker@tuckerd.info',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/tjdcs/specio',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<3.11',
}


setup(**setup_kwargs)



from dataclasses import dataclass
import re
from typing import List

import numpy as np
from numpy import ndarray
from specio.common import Measurement

from specio.protobuf import measurements_pb2

__author__ = "Tucker Downs"
__copyright__ = "Copyright 2022 Specio Developers"
__license__ = "MIT License - https://github.com/tjdcs/specio/blob/main/LICENSE.md"
__maintainer__ = "Tucker Downs"
__email__ = "tucker@tuckerd.info"
__status__ = "Development"

FILE_EXTENSION = ".csmf"


@dataclass
class MeasurementList_Notes:
    notes: None | str = None
    author: None | str = None
    location: None | str = None
    software: None | str = "colour-specio"


@dataclass
class MeasurementList:
    measurements: List[Measurement]
    metadata: MeasurementList_Notes = MeasurementList_Notes()
    test_colors: ndarray | None = None
    order: List[int] | None = None
    pass


def save_measurements(
    file: str,
    measurements: Measurement | List[Measurement],
    notes: MeasurementList_Notes = MeasurementList_Notes(),
    order: None | List[int] = None,
    testColors: None | ndarray | List[List[float]] = None,
):
    if type(measurements) == Measurement:
        measurements = [measurements]

    pbuf = measurements_pb2.Measurement_List()

    m: Measurement
    for m in measurements:
        m_pbuf = m.to_protobuf()[1]
        pbuf.measurements.append(m_pbuf)

    if notes.notes:
        pbuf.notes = notes.notes

    if notes.author:
        pbuf.author = notes.author

    if notes.location:
        pbuf.location = notes.location

    if notes.software:
        pbuf.software = notes.software

    if order is not None:
        pbuf.order[:] = order

    if testColors is not None:
        if type(testColors) == list:
            testColors = np.array(testColors)
        if np.ptp(testColors) > 1 and np.all(np.modf(testColors)[0] < 1e-8):
            for color in testColors:
                tc_buf = measurements_pb2.Measurement_List.TestColor()
                tc_buf.c[:] = [int(value) for value in color.tolist()]
                pbuf.colors.append(tc_buf)
        else:
            for color in testColors:
                tc_buf = measurements_pb2.Measurement_List.TestColor()
                tc_buf.f[:] = color
                pbuf.colors.append(tc_buf)

    data_string = pbuf.SerializeToString()

    with open(file=file + FILE_EXTENSION, mode="wb") as f:
        f.write(data_string)
    pass


def load_measurements(file: str) -> MeasurementList:
    data_string: bytes
    file = re.sub(".csmf", "", file)
    with open(file=file + FILE_EXTENSION, mode="rb") as f:
        data_string = f.read()

    pbuf = measurements_pb2.Measurement_List()
    pbuf.ParseFromString(data_string)

    measurements: List[Measurement] = []
    for mbuf in pbuf.measurements:
        measurements.append(Measurement(mbuf))

    measurements: MeasurementList = MeasurementList(
        measurements=measurements, order=pbuf.order, test_colors=pbuf.test_colors
    )

    return measurements


if __name__ == "__main__":
    from time import time

    def timer_func(func):
        # This function shows the execution time of
        # the function object passed
        def wrap_func(*args, **kwargs):
            t1 = time()
            result = func(*args, **kwargs)
            t2 = time()
            print(f"Function {func.__name__!r} executed in {(t2-t1):.4f}s")
            return result

        return wrap_func

    measures = []
    for i in range(100):
        measures.append(Measurement())

    file = "/Users/tucker/Downloads/test"

    @timer_func
    def time_save():
        save_measurements(file=file, measurements=measures)

    time_save()

    @timer_func
    def time_read():
        load_measurements(file=file)

    time_read()
